@@ -46,8 +46,6 @@ from sklearn.naive_bayes import MultinomialNB
 from sklearn.linear_model import LogisticRegression
 from sklearn import metrics
 
-
-
 #### For Regression problems
 from sklearn.model_selection import train_test_split
 from sklearn.pipeline import Pipeline
@@ -144,7 +142,21 @@ def left_subtract(l1,l2):
             lst.append(i)
     return lst
 ################################################################################
+import nltk
+def tokenize_and_stem(text):
+    stemmer = SnowballStemmer("english")
+    # first tokenize by sentence, then by word to ensure that punctuation is caught as it's own token
+    tokens = [word for sent in nltk.sent_tokenize(text) for word in nltk.word_tokenize(sent)]
+    filtered_tokens = []
+    # filter out any tokens not containing letters (e.g., numeric tokens, raw punctuation)
+    for token in tokens:
+        if re.search('[a-zA-Z]', token):
+            filtered_tokens.append(token)
+    stems = [stemmer.stem(t) for t in filtered_tokens]
+    return stems
+################################################################################
 from sklearn.feature_extraction import text 
+from nltk.stem.snowball import SnowballStemmer
 def select_best_nlp_vectorizer(model, data, col, target, metric,
                     seed, modeltype,min_df):
     """
@@ -157,7 +169,22 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
     ####   a  good idea, since it results in much cleaner and much better terms selected.
     ################################################################################
     """
-    stop_words = text.ENGLISH_STOP_WORDS
+    nltk.download("popular")
+    add_words = ["s"]
+    from sklearn.feature_extraction import text
+    #stopWords = text.ENGLISH_STOP_WORDS.union(add_words)
+    from nltk.corpus import stopwords
+    stopWords = set(stopwords.words('english')).union(add_words)
+    excl =['will',"i'll",'shall',"you'll",'may',"don't","hadn't","hasn't","haven't",
+           "don't","isn't",'if',"mightn't","mustn'","mightn't",'mightn',"needn't",
+           'needn',"needn't",'no','not','shan',"shan't",'shouldn',"shouldn't","wasn't",
+          'wasn','weren',"weren't",'won',"won't",'wouldn',"wouldn't","you'd",'you',
+          "you'd","you'll","you're",'yourself','yourselves']
+    stopWords = [x for x in stopWords if x not in excl]
+    ######################################################
+    #### This calculates based on the average number of words in an NLP column how many max_features
+    max_features = int(data[col].map(len).mean()*2)
+    ######################################################
     if len(data) >= 1000000:
         max_features = 1000
     elif len(data) >= 100000:
@@ -185,7 +212,7 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
     ################################################################
     if modeltype is None or modeltype == '':
         print('Since modeltype is None, Using TFIDF vectorizer with min_df and max_features')
-        tvec = TfidfVectorizer(ngram_range=(1,3), stop_words='english', max_features=int(max_features*0.5), min_df=min_df)
+        tvec = TfidfVectorizer(ngram_range=(1,3), stop_words=stopWords, max_features=int(max_features*0.5), min_df=min_df)
         data_dtm =  data[col]
         data_dtm = tvec.fit_transform(data_dtm)
         print('Features: ', data_dtm.shape[1])
@@ -204,7 +231,7 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
     for min_df in sorted(np.linspace(0.10,0.01,10)):
         try:
             vect_5000 = CountVectorizer(ngram_range=(1, 3), max_features=int(max_features*0.5),
-                                min_df=min_df, binary=False, stop_words='english')
+                                min_df=min_df, binary=False, stop_words=stopWords)
             all_vecs[vect_5000], all_models[vect_5000] = tokenize_test_by_metric(model, X_train, X_test, y_train,
                             y_test, target, metric,
                               vect_5000, seed, modeltype)
@@ -212,27 +239,51 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
             break
         except:
             continue
-    print('\n#### Using lemmas for feature extraction, with min_df and features (WARNING: SLOW!)')
-    vect_lemma = CountVectorizer(analyzer=split_into_lemmas, max_features=int(max_features*0.5), ngram_range=(1, 3), 
-                    min_df=min_df, binary=True,stop_words='english')
+    print('\n#### Using Vectorizer with max_features < 100 and a low 0.001 min_df with n_gram (2-5)')
+    ##########################################################################
+    ##### It's BEST to use small max_features (50) and a low 0.001 min_df with n_gram (2-5).
+    ######  There is no need in that case for stopwords or analyzer since the 2-grams take care of it
+    #### Once you do above, there is no difference between count_vectorizer and tfidf_vectorizer
+    #### Once u do above, increasing max_features from 50 to even 500 doesn't get you a higher score!
+    ##########################################################################
+    vect_lemma = CountVectorizer(
+                                   max_features=int(max_features*0.5), 
+                                   ngram_range=(2, 5), 
+                                    min_df=0.001, 
+                                   binary=True,
+                                    )
     try:
         all_vecs[vect_lemma], all_models[vect_lemma] = tokenize_test_by_metric(model, X_train, X_test, y_train,
                                            y_test, target, metric,
                                              vect_lemma, seed, modeltype)
     except:
-        print('Vectorizer with Lemma and Min_df skipped')
+        print('Error: Using CountVectorizer')
 
     print('\n# Using TFIDF vectorizer with min_df and max_features')
     if modeltype != 'Regression':
         tvec = TfidfVectorizer( max_features=int(max_features*0.5), 
-                                stop_words='english', ngram_range=(1, 3), min_df=min_df, binary=True)
+                                stop_words=stopWords, ngram_range=(1, 3), min_df=min_df, binary=True)
     else:
         tvec = TfidfVectorizer( max_features=int(max_features*0.5), 
-                                stop_words='english', ngram_range=(1, 3), min_df=min_df, binary=False)
+                                stop_words=stopWords, ngram_range=(1, 3), min_df=min_df, binary=False)
     all_vecs[tvec], all_models[tvec] = tokenize_test_by_metric(model, X_train, X_test, y_train,
                                       y_test, target, metric,
                                         tvec, seed, modeltype)
-    ######## Once you have built 3 different transformers it is time to compare them
+
+    print('\n# Using TFIDF vectorizer with Snowball Stemming and max_features')
+    if modeltype != 'Regression':
+        tvec2 = TfidfVectorizer(max_df=0.80, max_features=int(max_features*0.5),
+                                 min_df=min_df, stop_words=stopWords, binary=True,
+                                 use_idf=True, tokenizer=tokenize_and_stem, ngram_range=(1,3))
+    else:
+        tvec2 = TfidfVectorizer(max_df=0.80, max_features=int(max_features*0.5),
+                                 min_df=min_df, stop_words=stopWords, binary=False,
+                                 use_idf=True, tokenizer=tokenize_and_stem, ngram_range=(1,3))
+    all_vecs[tvec2], all_models[tvec2] = tokenize_test_by_metric(model, X_train, X_test, y_train,
+                                      y_test, target, metric,
+                                        tvec2, seed, modeltype)
+
+    ######## Once you have built 4 different transformers it is time to compare them
     if modeltype.endswith('Classification'):
         best_vec = pd.Series(all_vecs).idxmax()
     else:
@@ -246,7 +297,7 @@ def select_best_nlp_vectorizer(model, data, col, target, metric,
     return best_vec, all_models[best_vec], data_dtm, min_df
 
 ############################################################################
-from sklearn.metrics import balanced_accuracy_score
+from sklearn.metrics import balanced_accuracy_score,mean_absolute_error,mean_squared_error
 def return_scoreval(scoretype, y_true, y_preds, y_proba, modeltype):
     if modeltype.endswith('Classification'):
         if scoretype == 'f1':
@@ -352,8 +403,8 @@ def select_top_features_from_vectorizer(X, vectorizer,top_n=25):
         best_features_array = XA[:]
         ##### There are instances where dicti is empty or has only 1-grams. So test to make sure!
         if len(dicti) > 0:
-            for i in range(1,len(dicti)+1):
-                for eachterm in dicti[i]:
+            for i,key in enumerate(dicti):
+                for eachterm in dicti[key]:
                     if iteration == 1:
                         try:
                             index = vectorizer.vocabulary_[eachterm]
@@ -380,7 +431,7 @@ from xgboost.sklearn import XGBRegressor
 from sklearn.naive_bayes import GaussianNB
 import xgboost as xgb
 def Auto_NLP(nlp_column, train, test, target, score_type,
-                            seed, modeltype,top_num_features=25):
+                            seed, modeltype,top_num_features=25, verbose=0):
     """
     ##################################################################################
     #### Auto_NLP expects both train and test to be data frames with one NLP column 
@@ -479,6 +530,7 @@ def Auto_NLP(nlp_column, train, test, target, score_type,
     #### Now let's do a combined transformation of NLP column
     nlp_data, nlp_summary_cols = create_summary_of_nlp_cols(nlp_data, nlp_column)
     nlp_result_columns += nlp_summary_cols
+    print('    Added %d columns for counts of words and characters in each row' %len(nlp_summary_cols))
     #### next create parts-of-speech tagging ####
     if len(nlp_data) <= 10000:
         ### VADER is accurate but very SLOOOWWW. Do not do this for Large data sets ##############
@@ -607,7 +659,7 @@ def plot_histogram_probability(dist_train, dist_test, label_title):
     plt.show();
 ########################################################################
 module_type = 'Running' if  __name__ == "__main__" else 'Imported'
-version_number = '0.0.14'
+version_number = '0.0.16'
 print("""Imported Auto_NLP version: %s.. Call using:
      train_nlp, test_nlp, best_nlp_transformer = Auto_NLP(nlp_column, train, test, target, score_type, seed, modeltype)""" %version_number)
 ########################################################################
